module AMR.WithDenseList exposing (Model, decoder, encode, init, update)

import DenseList exposing (DenseList)
import Json.Decode
import Json.Encode


type alias Model model =
    { appModel : model
    , appMsgs : DenseList String
    }


init :
    (msg -> model -> ( model, Cmd msg ))
    -> Json.Decode.Decoder msg
    -> Maybe String
    -> ( model, Cmd msg )
    -> ( Model model, Cmd msg )
init updateApp msgDecoder fromCache ( appModel, appCmd ) =
    case Maybe.map (Json.Decode.decodeString (decoder (withoutCmd updateApp) msgDecoder appModel)) fromCache of
        Just (Ok model) ->
            ( model
            , Cmd.none
            )

        _ ->
            ( { appModel = appModel
              , appMsgs = DenseList.empty
              }
            , appCmd
            )


update :
    (msg -> model -> ( model, Cmd msg ))
    -> (String -> Cmd msg)
    -> (msg -> Json.Encode.Value)
    -> msg
    -> Model model
    -> ( Model model, Cmd msg )
update updateApp toCache encodeMsg appMsg model =
    let
        ( updateModel, appCmd ) =
            updateApp appMsg model.appModel

        updatedModel =
            { model
                | appModel = updateModel
                , appMsgs = DenseList.insert (Json.Encode.encode 0 (encodeMsg appMsg)) model.appMsgs
            }
    in
    ( updatedModel
    , Cmd.batch
        [ appCmd
        , toCache (Json.Encode.encode 0 (encode updatedModel))
        ]
    )


encode : Model model -> Json.Encode.Value
encode model =
    DenseList.encode Json.Encode.string model.appMsgs


decoder :
    (msg -> model -> model)
    -> Json.Decode.Decoder msg
    -> model
    -> Json.Decode.Decoder (Model model)
decoder replayUpdate msgDecoder model =
    Json.Decode.map
        (\( appMsgs, appModel ) ->
            { appModel = appModel
            , appMsgs = appMsgs
            }
        )
        (DenseList.foldLeftUntilErrorDecoder
            Json.Decode.string
            (Json.Decode.decodeString msgDecoder)
            replayUpdate
            model
        )


withoutCmd : (msg -> model -> ( model, Cmd msg )) -> msg -> model -> model
withoutCmd updateApp msg model =
    Tuple.first (updateApp msg model)
